#############################################################################
# Author: 
# Guoyan Zhao <gzhao@pathology.wustl.edu>
# ***************************************************************************
# Copyright (C) Washington University in St. Louis. Developed by Guoyan 
# Zhao, Washington University in St. Louis, Department of Pathology and 
# Immunology.
#
# This work is licensed under the Open Source License V2.1. To view a copy 
# of this license, visit http://www.opensource.org/licenses/osl-2.1.php  
#
# #############################################################################

#!/usr/bin/perl -w
use strict;

my $usage = ' This script is designed to find out how the 4959 motifs are 
distributed in each given promoter sequence and predicts modules based on 
the distribution.

perl CERMOD.pl  <sequence file><matrix dir><output choice>
<sequence file> = fasta format sequence file
<matrix dir> = full path to the directory that holds the 4959 motifs.
<output choice> = 0: only output Cis-regulatory module information
                  1: output Cis-regulatory module and Z-score distribution

output to standard output

';

die $usage unless @ARGV == 3;
my ($seq_f, $dir, $output_choice) = @ARGV;

my $HOME = $ENV{HOME};

my $Z_score_cutoff = 3.09; # corresponding to p vlaue = 1.0E-3
#my $Z_score_cutoff = 3.719; # corresponding to p vlaue = 1.0E-4
#my $Z_score_cutoff = 4.265; # corresponding to p vlaue = 1.0E-5
my $distance_cutoff = 30; # distance between the nearby peakes
my @gene_names = ();
my %seq = (); # gene name => sequence
&getGeneSeq($seq_f, \%seq, \@gene_names);
#print "finished geting gene seq\n";

##########################################################################
# Read matrices and descriptions in dir
my @names = (); # Names of the matrix file and desc file
my %width = (); # matrix name => width of matrix
&read_names( \@names, $dir ); # Read list of matrix names in dir
&get_width($dir, \@names, \%width);

# for each gene, find modules in it
my @sites = (); #  [all the sites for all matrices in this gene]}
my %MatrixSites = (); # motif name => [position1, position2, ...]
foreach my $gene (@gene_names) {
	my $seq_length = length($seq{$gene});
#	print "gene is $gene, seq length = $seq_length\n";

	@sites = (); #  [all the sites for all matrices in this gene]}
	%MatrixSites = (); # motif name => [position1, position2, ...]
	&get_allMotifSites(\@sites, \@names, $gene, $seq{$gene}, \%MatrixSites);

	my $modules_ref = &find_module($gene, $seq_length, \@sites);
	foreach my $mod (@{$modules_ref}) {
		print $mod, "\n";
		my %composit_matrix = ();
		my ($geneName, $module_start, $module_end) = split (/_/, $mod);
		foreach my $matrix (keys %MatrixSites) {
			foreach my $motif_position (@{$MatrixSites{$matrix}}) {
				if (($motif_position >= $module_start) && ($motif_position <= $module_end)) {
					$composit_matrix{$matrix} = 1;
#					print "$mod, composit matrix: $matrix, $motif_position\n";
				}
			}
		}

		# print out composition matrices
#			foreach my $key (keys %composit_matrix) {
#				print $key, " ";
#			}
#			print "\n";
	}
}

print "# program finished .\n";

exit;

################################################################
sub find_module() {
	my ($gene, $length, $sites_ref) = @_;

	# bin the positions, each position is a bin  
	my $binNum = $length;
	my @bins = ();
	for (my $i = 0; $i < $binNum; $i++) {
		$bins[$i] = 0;
	}

	foreach my $pos (@{$sites_ref}) {
		$bins[$pos]++;
	}

	# calculate average, SD
	my $sum = 0;
	for (my $i = 0; $i < $binNum; $i++) {
		$sum += $bins[$i];
	}
	my $ave = $sum/$binNum;
	my $variance = 0;
	for (my $i = 0; $i < $binNum; $i++) {
		$variance += ($bins[$i] - $ave)*($bins[$i] -  $ave);
	}
	$variance = $variance/$binNum;
	my $SD = sqrt($variance);
	#print "ave = $ave, SD = $SD\n";
	if ($SD == 0) {
		return;
	}


	if ($output_choice == 1) {
		print "Z score distribution along the sequence $gene:\n";
	}

	my @Z_score = ();
	for (my $i = 0; $i < $binNum; $i++) {
		$Z_score[$i] = ($bins[$i] - $ave)/$SD;
		#print "value = $bins[$i], z score = $Z_score[$i]\n";
	
#		print out Z score distribution
		if ($output_choice == 1) {
			print $i+1, "\t", $Z_score[$i], "\n";
		}
	}
	print "\n";

	# first find all the peaks that have z score >= cutoff
	my @peaks = ();
	for (my $i = 0; $i < (scalar @Z_score); $i++) {
		if ($Z_score[$i] >= $Z_score_cutoff) {
			push @peaks, $i;
		}
	}
	my %peak_use = (); # position => used 1, not used 0
	foreach my $peak_pos (@peaks) {
		$peak_use{$peak_pos} = 0;
	}

	# extend from each peak
	my $startPos = 0;
	my $lastPos = 0;
	my $distance = 0; # distance from the last peak being included to the next negative z score position 
	my @module = ();
	foreach my $peak_pos (@peaks) {
		my $startPos = $peak_pos;
		my $lastPos = $peak_pos;
		if (!$peak_use{$peak_pos}) {
			# extend backward 5'
			my $distance_start = $peak_pos; # start for calculate distance
			for (my $i = $peak_pos - 1; $i > 0; $i--) {
				if ($Z_score[$i] > 0) { # position has positive Z score
					$startPos = $i + 1; # sequence start position
					if ($Z_score[$i] >= $Z_score_cutoff) { # is a peak
						$peak_use{$i} = 1;
						$distance_start = $i;
					}
				}
				else { # position has negative Z score
					$distance = $distance_start - $i + 1;
					if ($distance > $distance_cutoff) {
						last;
					}
				}
			}

			# extend forward 3'
			for (my $i = $peak_pos + 1; $i < $binNum; $i++) {
				if ($Z_score[$i] > 0) { # position has positive Z score
					$lastPos = $i + 1; # sequence end position
					if ($Z_score[$i] >= $Z_score_cutoff) { # is a peak
						$peak_use{$i} = 1;
						$distance_start = $i;
					}
				}
				else { # position has negative positive Z score
					$distance = $i - $distance_start + 1;
					if ($distance > $distance_cutoff) {
						last;
					}
				}
			}
	
			my $mod_start = $startPos - 15; # subtract 15 bp to cover the possible longest motif
			if ($mod_start < 0) { $mod_start = 0;}
			my $mod_end = $lastPos + 15;  # add 15 bp to cover the possible longest motif
			if ($mod_end > $length) { $mod_end = $length;}

			my $mod = $gene."_".$mod_start."_".$mod_end;
			push @module, $mod;
		}
	}
	undef @bins;
	undef @Z_score;
=head1
	print "before procesing:\n";
	foreach my $mod (@module) {
		print $mod, "\n";
	}
=cut

	# merge modules if their ends are within 30 bp
	my @post_processed_modules = ();
	if (scalar @module) {
		my ($name, $start_i, $end_i) = split("_", $module[0]);
		my $new_start = $start_i;
		my $new_end = $end_i;
		my $i = 1;
		while( $i <= (scalar @module - 1)) {
			($name, $start_i, $end_i) = split ("_", $module[$i]);
			if ($start_i - $new_end <= 30) { 
				$new_end = $end_i;
				$i++;
			}
			else {
				my $p_modules = $gene."_".$new_start."_".$new_end;
				push @post_processed_modules, $p_modules;
	
				$new_start = $start_i;
				$new_end = $end_i;
				$i++;
			}
		}

		# for the last module
		my $p_module = $gene."_".$new_start."_".$new_end;
		push @post_processed_modules, $p_module;
	}

=head1	
	print "after procesing:\n";
	foreach my $mod (@post_processed_modules) {
		print $mod, "\n";
	}
	print "*****************\n\n";
=cut

	undef (@module);
	return \@post_processed_modules;
}


#####################################################################
sub get_allMotifSites() {
	my ($sites_ref, $names_ref, $gene_name, $seq, $MatrixSites_hash_ref) = @_;

	my $pat_f = "/tmp/temp_pat".$$;
	my $con_f = "/tmp/temp_con".$$;
	open( Con,">$con_f") or die "can notp open $con_f.\n";
	print Con "$gene_name \\ ";
	print Con $seq, "\n\\\n";
	close Con;

	my $matrix_name = "";
	foreach my $mat (@{$names_ref}) {
		# Run patser default cutoff and get sites for each motif
		my $command = $HOME."/tools/bin/patser_v3d_linux -a ".$HOME."/tools/bin/alphabet -m ".$dir."/".$mat.".matrix -f $con_f -c -d2 -li > $pat_f ";
		#print $command, "\n";
		system( $command );
	
		open ( PAT, $pat_f ) || die "Cannot open $pat_f.";
		while ( <PAT> ) {
#			print $_;

			if ( /File containing the matrix:.*short_mergeSites\/(.*)$/) {
				$matrix_name = $1;
#				print "matrix name = $matrix_name\n";
			}

			if ( /\s+(.*?)\s+position=\s*(\d+C*)\s*score=\s*([\d\.]+)/ ) {
#				print "matrix = $matrix_name, name = $1, pos = $2, score = $3\n";
				my $name = $1;
				my $pos = $2;
				if ($pos =~ /C/) {
					$pos =~ s/C//;
				}
		
				# use the center of motif
				my $cen = $pos + int($width{$mat}*0.5);
				push @{$sites_ref}, $cen;	
#				print "name = $name, pos = $cen\n";
				if (defined $MatrixSites_hash_ref->{$matrix_name}) {
					push @{$MatrixSites_hash_ref->{$matrix_name}}, $pos;
				}
				else {
					$MatrixSites_hash_ref->{$matrix_name} = [$pos];
				}
			}
		}
		close ( PAT );
	}
	unlink $pat_f;
	unlink $con_f;
}


################################################################3
sub read_names()
{
    my ( $name_r, $dir ) = @_;
    opendir (DH, $dir) or die "Cannot open $dir";
    foreach my $file (readdir DH) {
	if ($file =~ /matrix/) {
	    my $name = $file;
	    $name =~ s/\.matrix//;
	    #print "name = \\$name\\\n";
	    push @$name_r, $name;
	}
    }
    close DH;
}


#############################################################
sub array_has
{
  my ( $array_r, $ele ) = @_;
  for ( my $i=0; $i<=$#$array_r; $i++ )
  {
    if ( $ele eq $array_r->[$i] )
    {
      return 1;
    }
  }
  return 0;
}

#######################################################################
# read fasta file
sub getGeneSeq() {
    my ($fileName, $hashRef, $gene_name_hash) =  @_;
    open (File, $fileName) or die "Can not open file $fileName !\n";

    # get all the gene Name and seq name=> seq
    my $oldSeperator = $/;
    $/ = ">"; # seperates two records
    my $records = <File>;
    while ($records = <File>){
		#print $line;
		my @lines = split ("\n", $records);
		my $first_line = shift @lines;
		my @fields = split (" ", $first_line);
		my $name = shift @fields;
		my $seq = join ("", @lines);
		$seq =~ s/>//g;
		$seq =~ s/\s//g;
		#lc $seq;
 
#		print "name is \\$name\\\n";
#		print "seq is \n";
#		print "\\$seq\\\n";
		push @{$gene_name_hash}, $name;
		$hashRef->{$name} = $seq;
    }
    $/ = $oldSeperator;
	close File;
}

#############################################
sub get_width {
    my ($dir, $name_ref, $width_ref) = @_;

    my @line;
    my $desc;

    for ( my $i=0; $i < scalar @{$name_ref}; $i++ )
    {
        my $alignMatrix = "";
        my ($A, $C, $G, $T) = 0;

        $desc = $dir."/".$name_ref->[$i].".desc";
        open ( IN, $desc ) || die "Cannot open $desc.";
        while ( <IN> ) {
            if ( /width\s*=\s*(\d+)/ )   {
                $width_ref->{$name_ref->[$i]} = $1;
                last;
            }
        }
        close ( IN );
    }
}


